use axum::body::Body;
use axum::extract::MatchedPath;
use axum::http::{Request, StatusCode};
use axum::middleware::Next;
use axum::response::IntoResponse;

pub async fn print_request_response(
    request: Request<Body>,
    next: Next<Body>,
) -> Result<impl IntoResponse, (StatusCode, String)> {
    let matched_path = request
        .extensions()
        .get::<MatchedPath>()
        .map(MatchedPath::as_str);

    let span = tracing::info_span!(
        "http_request",
        method = ?request.method(),
        matched_path,
    );

    let span_ptr = span.enter();

    let (parts, body) = request.into_parts();

    let bytes = match hyper::body::to_bytes(body).await {
        Ok(bytes) => bytes,
        Err(err) => {
            return Err((
                StatusCode::BAD_REQUEST,
                format!("failed to read body: {}", err),
            ));
        }
    };

    tracing::info!(
        "Incoming request on '{} {}' with user agent '{}' and body: '{}'",
        parts.method,
        parts.uri,
        match parts.headers.get("user-agent") {
            Some(x) => x.to_str().unwrap_or("(failed to parse)"),
            None => "(empty)",
        },
        match std::str::from_utf8(&bytes) {
            Ok(body) => body,
            Err(e) => {
                tracing::warn!("Failed to transform request body to utf8: {:?}", e);
                ""
            }
        }
    );

    let res = next
        .run(Request::from_parts(parts, Body::from(bytes)))
        .await;

    let status = res.status();
    match status {
        StatusCode::OK | StatusCode::NOT_FOUND => {
            tracing::info!("request finished with status: {}", status)
        }
        _ => tracing::warn!("request finished with status: {}", status),
    }

    drop(span_ptr);

    Ok(res)
}
