use axum::extract::Host;
use axum::http::StatusCode;
use axum::response::IntoResponse;
use axum::Json;
use serde::{Deserialize, Serialize};
use std::fmt::{Display, Formatter};

#[derive(Serialize, Deserialize)]
struct DiscoveryDocument {
    pub authorization_endpoint: String,
    pub claims_supported: Vec<String>,
    pub code_challenge_methods_supported: Vec<String>,
    pub grant_types_supported: Vec<String>,
    pub id_token_signing_alg_values_supported: Vec<String>,
    pub issuer: String,
    pub jwks_uri: String,
    pub response_types_supported: Vec<String>,
    pub scopes_supported: Vec<String>,
    pub subject_types_supported: Vec<String>,
    pub token_endpoint: String,
    pub token_endpoint_auth_methods_supported: Vec<String>,
    pub userinfo_endpoint: String,
}

enum IssuerSegments {
    Authorize,
    Jwks,
    Token,
    Top,
    UserInfo,
}

impl Display for IssuerSegments {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self {
            IssuerSegments::Authorize => write!(f, "authorize"),
            IssuerSegments::Jwks => write!(f, "jwks"),
            IssuerSegments::Token => write!(f, "token"),
            IssuerSegments::Top => write!(f, ""),
            IssuerSegments::UserInfo => write!(f, "userinfo"),
        }
    }
}

fn format_issuer(host: &str, segment: IssuerSegments) -> String {
    match segment {
        IssuerSegments::Top => format!("https://{}/", host),
        _ => format!("https://{}/connect/{}", host, segment),
    }
}

pub async fn get(Host(host): Host) -> impl IntoResponse {
    tracing::info!("{}", host);
    let doc = DiscoveryDocument {
        authorization_endpoint: format_issuer(&host, IssuerSegments::Authorize),
        claims_supported: vec![
            "aud".to_string(),
            "email".to_string(),
            "exp".to_string(),
            "groups".to_string(),
            "iat".to_string(),
            "iss".to_string(),
            "sub".to_string(),
            "username".to_string(),
        ],
        code_challenge_methods_supported: vec!["S256".to_string(), "plain".to_string()],
        grant_types_supported: vec!["authorization_code".to_string(), "implicit".to_string()],
        id_token_signing_alg_values_supported: vec!["RS256".to_string()],
        issuer: format_issuer(&host, IssuerSegments::Top),
        jwks_uri: format_issuer(&host, IssuerSegments::Jwks),
        response_types_supported: vec![
            "code".to_string(),
            "code id_token".to_string(),
            "id_token".to_string(),
            "id_token token".to_string(),
        ],
        scopes_supported: vec!["email".to_string(), "openid".to_string()],
        subject_types_supported: vec!["public".to_string()],
        token_endpoint: format_issuer(&host, IssuerSegments::Token),
        token_endpoint_auth_methods_supported: vec![
            "client_secret_basic".to_string(),
            "client_secret_post".to_string(),
        ],
        userinfo_endpoint: format_issuer(&host, IssuerSegments::UserInfo),
    };

    (StatusCode::NOT_FOUND, Json(doc))
}
